==> ducene-0.4.0.tar.gz: lib/src/util/index_holder.dart <==
// Copyright (c) 2016, Shinichiro Abe. All rights reserved. Use of this source code
// is governed by a BSD-style license that can be found in the LICENSE file.

import 'dart:async';
import 'dart:io';

import '../../src/document/document.dart';
import '../../src/index/index_reader.dart';
import '../../src/store/index_directory.dart';
import '../../src/store/index_writer.dart';
import '../../src/search/index_searcher.dart';
import '../../src/search/query.dart';
import '../../src/store/sembast/sembast_index_directory.dart';
import '../../src/store/standard/composite_index_directory.dart';
import '../../src/store/standard/fs_index_directory.dart';
import '../../src/store/standard/ram_index_directory.dart';

/// An IndexHolder maintains multiple index segments.
///
/// It's best to use [DirectoryHolder.open(IndexHolderDirectory)] to obtain an IndexHolder.
class IndexHolder {
  final IndexHolderDirectory _holderDirectory;
  IndexReader _realTimeReader;

  IndexHolder(this._holderDirectory);

  Future<Null> init() async {
    await _holderDirectory.init();
    _realTimeReader = _holderDirectory.newRealTimeReader();
    await _realTimeReader.read();
  }

  Future<Null> updateDocuments(Iterable<Document> documents) async {
    final List<String> ids = <String>[];
    for (Document doc in documents) {
      ids.add(doc.get('id'));
    }
    for (String id in ids) {
      await deleteDocuments(new BoolQuery().append('id', id));
    }

    final NamedIndexDirectory nDir = _holderDirectory.newIndexDirectory();
    final IndexDirectory directory = nDir.directory;
    final IndexWriter writer = new IndexWriter(directory);
    final IndexTable table = writer.buildIndexTable(documents);
    await writer.writeTable(table);
    _holderDirectory.register(nDir);

    await _writeRealTime(table);
  }

  Future<Null> deleteDocuments(BoolQuery query) async {
    for (String name in _holderDirectory.directories.keys) {
      final IndexDirectory directory = _holderDirectory.directories[name];
      final IndexWriter writer = new IndexWriter(directory);
      await writer.delete(query);
    }

    if (_holderDirectory.directories.keys.length > 0) {
      await _deleteRealTime(query);
    }
  }

  Future<Null> _deleteRealTime(BoolQuery query) async {
    final IndexSearcher searcher = new IndexSearcher(_realTimeReader);
    final TopDocs hits = await searcher.search(
        query, await _realTimeReader.maxDoc(),
        scoreSort: ScoreSort.indexed());
    if (hits.totalHits > 0) {
      final Set<int> docIds =
          hits.scoreDocs.map((ScoreDoc scoreDoc) => scoreDoc.doc).toSet();
      final DocSet liveDocs = await _realTimeReader.liveDocs();
      final Set<int> newLiveDocs = liveDocs.toSet().difference(docIds);
      _realTimeReader.thisLiveDocs = new DocSet(newLiveDocs);
    }
  }

  Future<Null> _writeRealTime(IndexTable table) async {
    final int maxDoc = await _realTimeReader.maxDoc();

    final Set<int> currentDocIds = _realTimeReader.thisLiveDocs.toSet();
    final Set<int> liveDocs = table.liveDocs;
    CompositeLiveDocsStreams.merge(currentDocIds, liveDocs, maxDoc);
    _realTimeReader.thisLiveDocs.update(currentDocIds);

    final Map<String, Map<String, DocSet>> indexes =
        <String, Map<String, DocSet>>{};
    for (String field in table.indexes.keys) {
      indexes[field] = table.indexes[field].build();
    }
    CompositeIndexStreams.merge(_realTimeReader.thisIndexes, indexes, maxDoc);

    final Map<String, Map<int, DocData>> data = <String, Map<int, DocData>>{};
    for (String field in table.data.keys) {
      data[field] = table.data[field].build();
    }
    CompositeDataStreams.merge(_realTimeReader.thisData, data, maxDoc);
  }

  Future<Null> forceMerge() async {
    final NamedIndexDirectory nDir = _holderDirectory.newIndexDirectory();
    final IndexDirectory directory = nDir.directory;
    await IndexWriter.copyFrom(_newCompositeIndexDirectory(), directory);
    final IndexWriter writer = new IndexWriter(directory);
    await writer.expungeDeletes();
    await _holderDirectory.release();
    _holderDirectory.register(nDir);

    _forceMergeRealTime();
  }

  void _forceMergeRealTime() {
    final Map<String, Map<String, DocSet>> newIndexes =
        <String, Map<String, DocSet>>{};
    IndexWriter.expungeDeletesForIndexes(
        _realTimeReader.thisIndexes, newIndexes, _realTimeReader.thisLiveDocs);
    _realTimeReader.thisIndexes.clear();
    for (String field in newIndexes.keys) {
      _realTimeReader.thisIndexes[field] = newIndexes[field];
    }

    final Map<String, Map<int, DocData>> newData =
        <String, Map<int, DocData>>{};
    IndexWriter.expungeDeletesForData(
        _realTimeReader.thisData, newData, _realTimeReader.thisLiveDocs);
    _realTimeReader.thisData.clear();
    for (String field in newData.keys) {
      _realTimeReader.thisData[field] = newData[field];
    }
  }

  CompositeIndexDirectory _newCompositeIndexDirectory() {
    return new CompositeIndexDirectory(
        _holderDirectory.directories.values.toList());
  }

  Future<IndexReader> openIndexReader() async {
    final IndexReader ir =
        await DirectoryReader.open(_newCompositeIndexDirectory());
    return ir;
  }

  IndexReader openRealTimeIndexReader() {
    return _realTimeReader;
  }

  Future<IndexSearcher> newIndexSearcher() async {
    final IndexReader reader = await openIndexReader();
    final IndexSearcher searcher = new IndexSearcher(reader);
    return searcher;
  }

  IndexSearcher newRealTimeIndexSearcher() {
    final IndexReader reader = openRealTimeIndexReader();
    final IndexSearcher searcher = new IndexSearcher(reader);
    return searcher;
  }
}

abstract class IndexHolderDirectory {
  Map<String, IndexDirectory> directories = <String, IndexDirectory>{};
  Future<Null> init();
  IndexReader newRealTimeReader();
  NamedIndexDirectory newIndexDirectory();
  Future<Null> release();
  void register(NamedIndexDirectory directoryKeyValue) {
    directories[directoryKeyValue.name] = directoryKeyValue.directory;
  }
}

class DirectoryHolder {
  static Future<IndexHolder> open(IndexHolderDirectory directoryHolder) async {
    final IndexHolder holder = new IndexHolder(directoryHolder);
    await holder.init();
    return holder;
  }
}

class NamedIndexDirectory {
  final String name;
  final IndexDirectory directory;
  NamedIndexDirectory(this.name, this.directory);
}

// [RAMIndexDirectory] based implementation.
class RAMIndexHolderDirectory extends IndexHolderDirectory {
  @override
  Future<Null> init() {
    return null;
  }

  @override
  IndexReader newRealTimeReader() {
    return new IndexReader(new RAMIndexDirectory());
  }

  @override
  NamedIndexDirectory newIndexDirectory() {
    final String name = (directories.length + 1)
        .toString()
        .padLeft(10, '0'); //.toRadixString(36);
    final IndexDirectory directory = new RAMIndexDirectory();
    return new NamedIndexDirectory(name, directory);
  }

  @override
  Future<Null> release() async {
    directories.clear();
  }
}

/// [FSIndexDirectory] based implementation.
class FSIndexHolderDirectory extends IndexHolderDirectory {
  final Directory _dir;
  FSIndexHolderDirectory(this._dir);

  @override
  Future<Null> init() {
    if (_dir.existsSync()) {
      final List<String> paths = _dir
          .listSync(recursive: false, followLinks: false)
          .map((FileSystemEntity e) => e.path)
          .toList();
      for (String path in paths) {
        final String name = path.split('/').last;
        final IndexDirectory directory =
            new FSIndexDirectory(new Directory(path));
        directories[name] = directory;
      }
    }
    return null;
  }

  @override
  IndexReader newRealTimeReader() {
    return new IndexReader(
        new CompositeIndexDirectory(directories.values.toList()));
  }

  @override
  NamedIndexDirectory newIndexDirectory() {
    int number;
    if (directories.length == 0) {
      number = 0;
    } else {
      number = int.parse(directories.keys.last);
    }
    final String name = (number + 1).toString().padLeft(10, '0');
    final String path = '${_dir.path}/$name';
    final IndexDirectory directory = new FSIndexDirectory(new Directory(path));
    return new NamedIndexDirectory(name, directory);
  }

  @override
  Future<Null> release() async {
    for (String name in directories.keys) {
      final FSIndexDirectory directory = directories[name];
      final Directory target = directory.dir;
      target.deleteSync(recursive: true);
    }
    directories.clear();
  }
}

/// [SembastIndexDirectory] based implementation.
class SembastIndexHolderDirectory extends IndexHolderDirectory {
  final Directory _dir;
  final List<String> _indexedFields;
  final List<String> _storedFields;
  SembastIndexHolderDirectory(
      this._dir, this._indexedFields, this._storedFields);

  @override
  Future<Null> init() {
    if (_dir.existsSync()) {
      final List<String> paths = _dir
          .listSync(recursive: false, followLinks: false)
          .map((FileSystemEntity e) => e.path)
          .toList();
      for (String path in paths) {
        final String name = path.split('/').last;
        final IndexDirectory directory =
            new SembastIndexDirectory(path, _indexedFields, _storedFields);
        directories[name] = directory;
      }
    }
    return null;
  }

  @override
  IndexReader newRealTimeReader() {
    return new IndexReader(
        new CompositeIndexDirectory(directories.values.toList()));
  }

  @override
  NamedIndexDirectory newIndexDirectory() {
    int number;
    if (directories.length == 0) {
      number = 0;
    } else {
      number = int.parse(directories.keys.last);
    }
    final String name = (number + 1).toString().padLeft(10, '0');
    final String path = '${_dir.path}/$name';
    final IndexDirectory directory =
        new SembastIndexDirectory(path, _indexedFields, _storedFields);
    return new NamedIndexDirectory(name, directory);
  }

  @override
  Future<Null> release() async {
    for (String name in directories.keys) {
      final SembastIndexDirectory directory = directories[name];
      final Directory target = new Directory(directory.dbPath);
      target.deleteSync(recursive: true);
    }
    directories.clear();
  }
}

==> ducene-0.5.0.tar.gz: lib/src/util/index_holder.dart <==
// Copyright (c) 2016, Shinichiro Abe. All rights reserved. Use of this source code
// is governed by a BSD-style license that can be found in the LICENSE file.

import 'dart:async';
import 'dart:io';

import '../../src/index/document.dart';
import '../../src/index/index_reader.dart';
import '../../src/index/index_writer.dart';
import '../../src/search/index_searcher.dart';
import '../../src/search/query.dart';
import '../../src/store/index_directory.dart';
import '../../src/store/sembast/sembast_index_directory.dart';
import '../../src/store/standard/composite_index_directory.dart';
import '../../src/store/standard/fs_index_directory.dart';
import '../../src/store/standard/ram_index_directory.dart';

/// An IndexHolder maintains multiple index segments.
///
/// It's best to use [DirectoryHolder.open(IndexHolderDirectory)] to obtain an IndexHolder.
class IndexHolder {
  final IndexHolderDirectory _holderDirectory;
  IndexReader _realTimeReader;

  IndexHolder(this._holderDirectory);

  Future<Null> init() async {
    await _holderDirectory.init();
    _realTimeReader = _holderDirectory.newRealTimeReader();
    await _realTimeReader.read();
  }

  Future<Null> updateDocuments(Iterable<Document> documents) async {
    final List<String> ids = <String>[];
    for (Document doc in documents) {
      ids.add(doc.get('id'));
    }
    for (String id in ids) {
      await deleteDocuments(new BoolQuery().append('id', id));
    }

    final NamedIndexDirectory nDir = _holderDirectory.newIndexDirectory();
    final IndexDirectory directory = nDir.directory;
    final IndexWriter writer = new IndexWriter(directory);
    final IndexWriteModel model = writer.buildModel(documents);
    await writer.writeModel(model);
    _holderDirectory.register(nDir);

    await _writeRealTime(model);
  }

  Future<Null> deleteDocuments(BoolQuery query) async {
    for (String name in _holderDirectory.directories.keys) {
      final IndexDirectory directory = _holderDirectory.directories[name];
      final IndexWriter writer = new IndexWriter(directory);
      await writer.delete(query);
    }

    if (_holderDirectory.directories.keys.length > 0) {
      await _deleteRealTime(query);
    }
  }

  Future<Null> _deleteRealTime(BoolQuery query) async {
    final IndexSearcher searcher = new IndexSearcher(_realTimeReader);
    final TopDocs hits = await searcher.search(
        query, await _realTimeReader.maxDoc(),
        scoreSort: new NoOpScoreSort());
    if (hits.totalHits > 0) {
      final IDocSet liveDocs = await _realTimeReader.liveDocs();
      final IDocSet newLiveDocs = liveDocs.not(hits.docSet);
      _realTimeReader.liveDocsCache = IDocSet.newDocSetFromDocSet(newLiveDocs);
    }
  }

  Future<Null> _writeRealTime(IndexWriteModel model) async {
    final int maxDoc = await _realTimeReader.maxDoc();

    final Set<int> currentDocIds =
        _realTimeReader.liveDocsCache.toIterable().toSet();
    final Set<int> liveDocs = model.liveDocs.build().toIterable().toSet();
    CompositeLiveDocsStreams.merge(currentDocIds, liveDocs, maxDoc);
    final int latestMaxDoc = maxDoc + liveDocs.length;
    _realTimeReader.liveDocsCache.updateFromSet(currentDocIds, latestMaxDoc);

    _realTimeReader.indexInfoCache.setMaxDoc(latestMaxDoc);

    final Index indexes = Index.empty();
    for (String field in model.indexes.keys) {
      indexes[field] = model.indexes[field].build();
    }
    CompositeIndexStreams.merge(
        _realTimeReader.indexesCache, indexes, maxDoc, liveDocs.length);
    CompositeIndexStreams.resize(_realTimeReader.indexesCache,
        _realTimeReader.indexesCache, latestMaxDoc);

    final FieldData data = FieldData.empty();
    for (String field in model.data.keys) {
      data[field] = model.data[field].build();
    }
    CompositeDataStreams.merge(_realTimeReader.dataCache, data, maxDoc);
  }

  Future<Null> forceMerge() async {
    final NamedIndexDirectory nDir = _holderDirectory.newIndexDirectory();
    final IndexDirectory directory = nDir.directory;
    await IndexWriter.copyFrom(_newCompositeIndexDirectory(), directory);
    final IndexWriter writer = new IndexWriter(directory);
    await writer.expungeDeletes();
    await _holderDirectory.release();
    _holderDirectory.register(nDir);

    _forceMergeRealTime();
  }

  void _forceMergeRealTime() {
    //TODO see writer.expungeDeletes()
    final Index newIndexes = Index.empty();
    IndexWriter.expungeDeletesForIndexes(_realTimeReader.indexesCache,
        newIndexes, _realTimeReader.liveDocsCache);
    _realTimeReader.indexesCache.clear();
    for (String field in newIndexes.keys) {
      _realTimeReader.indexesCache[field] = newIndexes[field];
    }

    final FieldData newData = FieldData.empty();
    IndexWriter.expungeDeletesForData(
        _realTimeReader.dataCache, newData, _realTimeReader.liveDocsCache);
    _realTimeReader.dataCache.clear();
    for (String field in newData.keys) {
      _realTimeReader.dataCache[field] = newData[field];
    }
  }

  CompositeIndexDirectory _newCompositeIndexDirectory() {
    return new CompositeIndexDirectory(
        _holderDirectory.directories.values.toList());
  }

  Future<IndexReader> openIndexReader() async {
    final IndexReader ir =
        await DirectoryReader.open(_newCompositeIndexDirectory());
    return ir;
  }

  IndexReader openRealTimeIndexReader() {
    return _realTimeReader;
  }

  Future<IndexSearcher> newIndexSearcher() async {
    final IndexReader reader = await openIndexReader();
    final IndexSearcher searcher = new IndexSearcher(reader);
    return searcher;
  }

  IndexSearcher newRealTimeIndexSearcher() {
    final IndexReader reader = openRealTimeIndexReader();
    final IndexSearcher searcher = new IndexSearcher(reader);
    return searcher;
  }
}

abstract class IndexHolderDirectory {
  Map<String, IndexDirectory> directories = <String, IndexDirectory>{};
  Future<Null> init();
  IndexReader newRealTimeReader();
  NamedIndexDirectory newIndexDirectory();
  Future<Null> release();
  void register(NamedIndexDirectory directoryKeyValue) {
    directories[directoryKeyValue.name] = directoryKeyValue.directory;
  }
}

class DirectoryHolder {
  static Future<IndexHolder> open(IndexHolderDirectory directoryHolder) async {
    final IndexHolder holder = new IndexHolder(directoryHolder);
    await holder.init();
    return holder;
  }
}

class NamedIndexDirectory {
  final String name;
  final IndexDirectory directory;
  NamedIndexDirectory(this.name, this.directory);
}

// [RAMIndexDirectory] based implementation.
class RAMIndexHolderDirectory extends IndexHolderDirectory {
  @override
  Future<Null> init() {
    return null;
  }

  @override
  IndexReader newRealTimeReader() {
    return new IndexReader(new RAMIndexDirectory());
  }

  @override
  NamedIndexDirectory newIndexDirectory() {
    final String name = (directories.length + 1)
        .toString()
        .padLeft(10, '0'); //.toRadixString(36);
    final IndexDirectory directory = new RAMIndexDirectory();
    return new NamedIndexDirectory(name, directory);
  }

  @override
  Future<Null> release() async {
    directories.clear();
  }
}

/// [FSIndexDirectory] based implementation.
class FSIndexHolderDirectory extends IndexHolderDirectory {
  final Directory _dir;
  FSIndexHolderDirectory(this._dir);

  @override
  Future<Null> init() async {
    if (await _dir.exists()) {
      final Stream<String> paths = _dir
          .list(recursive: false, followLinks: false)
          .map((FileSystemEntity e) => e.path);
      await for (String path in paths) {
        final String name = path.split('/').last;
        final IndexDirectory directory =
            new FSIndexDirectory(new Directory(path));
        directories[name] = directory;
      }
    }
    return null;
  }

  @override
  IndexReader newRealTimeReader() {
    return new IndexReader(
        new CompositeIndexDirectory(directories.values.toList()));
  }

  @override
  NamedIndexDirectory newIndexDirectory() {
    int number;
    if (directories.length == 0) {
      number = 0;
    } else {
      number = int.parse(directories.keys.last);
    }
    final String name = (number + 1).toString().padLeft(10, '0');
    final String path = '${_dir.path}/$name';
    final IndexDirectory directory = new FSIndexDirectory(new Directory(path));
    return new NamedIndexDirectory(name, directory);
  }

  @override
  Future<Null> release() async {
    for (String name in directories.keys) {
      final FSIndexDirectory directory = directories[name];
      final Directory target = directory.dir;
      await target.delete(recursive: true);
    }
    directories.clear();
  }
}

/// [SembastIndexDirectory] based implementation.
class SembastIndexHolderDirectory extends IndexHolderDirectory {
  final Directory _dir;
  final List<String> _indexedFields;
  final List<String> _storedFields;
  SembastIndexHolderDirectory(
      this._dir, this._indexedFields, this._storedFields);

  @override
  Future<Null> init() async {
    if (await _dir.exists()) {
      final Stream<String> paths = _dir
          .list(recursive: false, followLinks: false)
          .map((FileSystemEntity e) => e.path);
      await for (String path in paths) {
        final String name = path.split('/').last;
        final IndexDirectory directory =
            new SembastIndexDirectory(path, _indexedFields, _storedFields);
        directories[name] = directory;
      }
    }
    return null;
  }

  @override
  IndexReader newRealTimeReader() {
    return new IndexReader(
        new CompositeIndexDirectory(directories.values.toList()));
  }

  @override
  NamedIndexDirectory newIndexDirectory() {
    int number;
    if (directories.length == 0) {
      number = 0;
    } else {
      number = int.parse(directories.keys.last);
    }
    final String name = (number + 1).toString().padLeft(10, '0');
    final String path = '${_dir.path}/$name';
    final IndexDirectory directory =
        new SembastIndexDirectory(path, _indexedFields, _storedFields);
    return new NamedIndexDirectory(name, directory);
  }

  @override
  Future<Null> release() async {
    for (String name in directories.keys) {
      final SembastIndexDirectory directory = directories[name];
      final Directory target = new Directory(directory.dbPath);
      await target.delete(recursive: true);
    }
    directories.clear();
  }
}
